import docker
import logging
from slugify import slugify
from pathlib import Path
from urllib.parse import urlparse
from pathlib import Path
from datetime import datetime
import pytz

logger = logging.getLogger(__name__)

import docker


def get_last_path_part(url):
    """
    Parses a URL and returns the last part of the path.
    
    Args:
        url (str): The URL to parse.
    
    Returns:
        str: The last part of the path in the URL.
    """
    # Parse the URL
    parsed_url = urlparse(url)
    # Extract the path and get the last part using pathlib
    last_part = Path(parsed_url.path).name
    return last_part

def ensure_network_exists(client, network_name, is_external=False, network_type="bridge"):
    """
    Ensure a Docker network exists.
    
    Args:
        network_name (str): Name of the network to ensure.
        is_external (bool): If True, assumes the network is managed externally and only verifies its existence.
        network_type (str): The type of the network ('bridge' or 'overlay'). Defaults to 'bridge'.
    
    Returns:
        docker.models.networks.Network: The existing or newly created network object.
    """
    try:
        # Initialize Docker client
     

        # Check if the network already exists
        existing_networks = client.networks.list(names=[network_name])
        if existing_networks:
            logger.debug(f"Network '{network_name}' already exists.")
            return existing_networks[0]

        if is_external:
            raise ValueError(f"External network '{network_name}' does not exist.")

        # Create the network if it doesn't exist
        network = client.networks.create(
            name=network_name,
            driver=network_type,
            check_duplicate=True
        )
        logger.debug(f"Network '{network_name}' created as a '{network_type}' network.")
        return network

    except docker.errors.APIError as e:
        logger.error(f"Error interacting with Docker API: {e}")
    except Exception as e:
        logger.error(f"Unexpected error: {e}")

    return None


def create_container(
    client: docker.DockerClient,
    image: str,
    local_dir: str = None,
    name: str = None,
    env_vars: dict = {},
    ports: dict = None,
    networks: list = None,
    labels: dict = None,
    remove=False
):
    """
    Create and start a Docker container for a development environment.

    :param image: The Docker image to use.
    :param local_dir: The local directory to bind to /workspace.
    :param env_vars: A dictionary of environment variables to set in the container.
    :param container_name: The name of the container.
    :param ports: A dictionary mapping container ports to host ports, e.g., {"8080": 8080}.
    :param networks: A list of networks to connect the container to.
    """
   
    
    # Prepare port bindings
    if isinstance(ports, dict):
        port_bindings = {f"{port}/tcp": host_port for port, host_port in (ports or {}).items()}
    elif isinstance(ports, list):
        port_bindings = {f"{port}/tcp": None for port in (ports or [])}
    else:
        port_bindings = None

    logger.info(f"Creating container from image '{image}'...")
    logger.debug(f"Port bindings: {port_bindings}")
    logger.debug(f"Volumes: {local_dir}")
    

    # Create and start the container
    container = client.containers.run(
        image=image,
        name=slugify(name),
        ports=port_bindings,
        environment=env_vars,
        volumes={local_dir: {"bind": "/workspace", "mode": "rw"}} if local_dir else None,
        detach=True,
        remove=remove, 
        labels=labels,
        network_mode="bridge",  # Can be overridden by connecting to networks after creation
    )
    
    # Connect to additional networks if specified
    for net in (networks or []):
        network = client.networks.get(net)
        logger.info(f"Connecting container to network '{net}' : {network.id}")
        network.connect(container)
    
    logger.info(f"Container '{container.name}' ({container.id}) created successfully.")
    return container



def only_one_container(containers, reset = False):
    if containers:
        logger.debug(f"Found {len(containers)} containers")
        first, rest = containers[0], containers[1:]
        
        for container in rest:
            logger.debug(f"Removing container {container.name}")
            container.remove(force=True)
        
        #if first is stopped, start it. 
        if reset:
            logger.debug(f"Destroying container {first.name} because reset is True")
            first.remove(force=True)
    
        elif first.status == "exited":
            logger.debug(f"Starting container {first.name}")
            first.start()
            return first
        else:
            logger.debug(f"Container {first.name} is already running")
            return first
        
    return None

def create_novnc_container(client, config,  username, reset = False):
    # Create the container
    
    containers = client.containers.list(filters={"label": f"jtl.novnc.username={username}"}, all=True)
    
    if container := only_one_container(containers, reset):
        return container
     
    name = slugify(username)
    container_name = f"{name}-novnc"
    hostname = f"{container_name}.do.jointheleague.org"
     
    labels = {
        "jtl":  'true', 
        "jtl.novnc": 'true', 
        "jtl.novnc.username": username,
        
        "caddy": hostname,
        "caddy.@ws.0_header": "Connection *Upgrade*",
        "caddy.@ws.1_header": "Upgrade websocket",
        "caddy.0_reverse_proxy": "@ws {{upstreams 6080}}",
        "caddy.1_reverse_proxy": "{{upstreams 6080}}"
    }
    
    container = create_container(
        client,
        image=config.IMAGES_NOVNC,
        name=container_name,
        #ports=["6080"],
        labels=labels,
        networks=["x11", "jtlctl", "caddy"],
    )

    # Start the container
    container.start()

    return container

def create_cs_container(client, config, image, username, env_vars, vnc_id=None, reset=False):
    # Create the container
    
    containers = client.containers.list(filters={"label": f"jtl.codeserver.username={username}"}, all=True)
    
    if container := only_one_container(containers, reset):
        return container
       
    vnc_c = client.containers.get(vnc_id)   
       
    name = slugify(username)
    container_name = f"{slugify(username)}"
       
    vnc_host = vnc_c.labels.get("caddy")
    vnc_url = f"https://{vnc_host}"

    password = "code4life"

    _env_vars = {
        "PASSWORD": password,
        "DISPLAY": f"{vnc_c.name}:0",
        "VNC_URL": vnc_url,
        "KST_REPORTING_URL": config.KST_REPORTING_URL,
        "KST_CONTAINER_ID": name,
		"KST_REPORT_RATE": config.KST_REPORT_RATE if hasattr(config, "KST_REPORT_RATE") else 30
    }
    
    env_vars = {**_env_vars, **env_vars}
    
    labels = {
        "jtl": 'true', 
        "jtl.codeserver": 'true',  
        "jtl.codeserver.username": username,
        "jt.codeserver.password": password,
        "jtl.codeserver.vnc": vnc_c.name,
        "jtl.codeserver.start_time": datetime.now(pytz.timezone('America/Los_Angeles')).isoformat(),
        
        "caddy": config.HOSTNAME_TEMPLATE.format(username=slugify(username)),
        "caddy.reverse_proxy": "{{upstreams 8080}}"
    }
    
    container = create_container(
        client,
        image=image,
        env_vars=env_vars,
        name=container_name,
        local_dir=None,
        ports = ["8080"],
        labels=labels,
        networks=["x11", "jtlctl", "caddy"],
    )

    # Start the container
    container.start()

    return container

def create_cs_pair(client, config, image, username, env_vars = {}, reset = False):
    """ Create a pair of containers: a novnc container and a codeserver container.
    
    """
    nvc = create_novnc_container(client, config, username=username, reset=reset)
    pa = create_cs_container(client, config, image, username=username, env_vars = env_vars, vnc_id=nvc.id, reset=reset)
    return nvc, pa


def container_list(client, all=False):
    containers = client.containers.list(filters={"label": f"jtl.codeserver"}, all=all)
    return containers

def container_status(client, username):
    containers = client.containers.list(filters={"label": f"jtl.codeserver.username={username}"}, all=True)
    
    if containers:
        return containers[0].status
    
    return 'non-exist'



from datetime import datetime, timedelta, timezone
from dateutil.relativedelta import relativedelta

def start_of_today():
    """Return the start of today as a string in isoformat in the Z timezone"""
   

    # Get current time
    now = datetime.now()

    # Get start of the day
    return datetime(now.year, now.month, now.day)

def start_of_yesterday():
    """Return the start of yesterday """
    

    # Get current time
    now = datetime.now()

    # Get start of yesterday
    return datetime(now.year, now.month, now.day) + timedelta(days=-1)


def start_of_tomorrow():
    """Return the start of tomorrow """
    

    # Get current time
    now = datetime.now()

    # Get start of tomorrow
    return datetime(now.year, now.month, now.day) + timedelta(days=1)

def end_of_today():
    """Return the end of today as a string in isoformat in the Z timezone"""
   

    # Get current time
    now = datetime.now()

    # Get end of the day
    return datetime(now.year, now.month, now.day) + timedelta(days=1, seconds=-1)

def end_of_yesterday():
    """Return the end of yesterday """
    

    # Get current time
    now = datetime.now()

    # Get end of yesterday
    return datetime(now.year, now.month, now.day) + timedelta(seconds=-1)

def start_of_day(dt: datetime):
    """Return the start of the day for the given datetime"""
    return dt.replace(hour=0, minute=0, second=0, microsecond=0)

def start_of_next_day(dt: datetime):
    """Return the start of the next day for the given datetime"""
    return start_of_day(dt) + timedelta(days=1)





def one_month_ago():
    return (datetime.now() - relativedelta(months=1)) \
        .replace(hour=0, minute=0, second=0, microsecond=0)


def one_month_before(dt: datetime):
    return (dt - relativedelta(months=1)).replace(hour=0, minute=0, second=0, microsecond=0)


def one_month_after(dt: datetime):
    return (dt + relativedelta(months=1)).replace(hour=23, minute=59, second=59, microsecond=0)

def add_months(dt: datetime, months: int):
    return (dt + relativedelta(months=months))

def month_range(start_date, end_date):
    """Generate Month ranges from the start date to the end date """
    while start_date <= end_date:
        yield start_date.replace(hour=0, minute=0, second=0, microsecond=0), \
            start_date + relativedelta(months=1)
        start_date += relativedelta(months=1)


def convert_naive(dt: datetime):
    """Convert a naive datetime to a timezone aware datetime, but only if it is naive"""

    if dt.tzinfo is None:
        from pytz import utc
        return utc.localize(dt)
    else:
        return dt





from datetime import datetime, timedelta
from src.jtlutil.config import * 
from pathlib import Path

from test import tree1, tree2

t1d = Path(tree1.__file__).parent
t2d = Path(tree2.__file__).parent

def test_find_config_file():
    
    # Build the dirs on the tree1 directory as root
    t1_config_dirs = get_config_dirs(root=t1d, home=t1d/'home'/'testuser', 
                                     cwd=t1d/'dev/l1/l2/l3/dev/')
    
    
    for d in t1_config_dirs:
        d.mkdir(parents=True, exist_ok=True)
    
    def strip_local(d):
        return str(d).rsplit('jtlutil/test/tree1', 1)[1]
    
    # Strip it back off. 
    d = [ strip_local(e) for e in t1_config_dirs]
    
    assert(d == ['/dev/l1/l2/l3/dev', '/home/testuser/.jtl', 
                '/etc/jtl', '/etc/jtl/secrets', '/app/config', 
                '/app/secrets', '/dev/l1/l2/l3/dev/secrets', 
                '/dev/l1/l2/l3/secrets'])
    
    dirs = [
        '/dev/l1/l2/l3/dev', 
        '/home/testuser/.jtl', 
        '/etc/jtl',
        '/etc/jtl/secrets', 
        '/app/config', 
        '/app/secrets', 
        '/dev/l1/l2/l3/dev/secrets', 
        '/dev/l1/l2/l3/secrets'
    ]
    
    config_files = ['config.env', 'dev.env', 'missing.env']

    for d in t1_config_dirs:
        for cf in config_files:
            tf = Path(d)/cf

            if cf != 'missing.env':
                tf.touch()
          
            try:
                p = find_config_file(cf, dirs=t1_config_dirs)
                assert p == tf, f"Expected {tf}, but got {p}"
            except FileNotFoundError as e:
                assert tf.name == 'missing.env', f"Expected missing file, but got {e}"
        
            if tf.exists():
                tf.unlink()
                

def test_config_object():
    config_data = {
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3'
    }
    
    config = Config(config_data)
    
    # Test attribute access
    assert config.key1 == 'value1'
    assert config.key2 == 'value2'
    assert config.key3 == 'value3'
    
    # Test item access
    assert config['key1'] == 'value1'
    assert config['key2'] == 'value2'
    assert config['key3'] == 'value3'
    
    # Test item setting
    config['key4'] = 'value4'
    assert config.key4 == 'value4'
    
    # Test item deletion
    del config['key4']
    try:
        _ = config.key4
        assert False, "Expected AttributeError"
    except AttributeError:
        pass
    
    # Test contains
    assert 'key1' in config
    assert 'key4' not in config
    
    # Test get method
    assert config.get('key1') == 'value1'
    assert config.get('key4') is None
    assert config.get('key4', 'default') == 'default'
    
    # Test keys, values, items
    assert set(config.keys()) == {'key1', 'key2', 'key3'}
    assert set(config.values()) == {'value1', 'value2', 'value3'}
    assert set(config.items()) == {('key1', 'value1'), ('key2', 'value2'), ('key3', 'value3')}
    
    config.key4 = 'value4'
    assert config['key4'] == 'value4'
    assert set(config.keys()) == {'key1', 'key2', 'key3', 'key4'}
    assert set(config.values()) == {'value1', 'value2', 'value3', 'value4'}
    
    
# Run tests with output: 
# pytest -s test/test_config.py::test_multi_config
def test_multi_config():
    
    import pprint
    
    d = {
    "DEVEL": "1",
    "PRODUCTION": '',
    "CONFIG": "CONFIG",
    "DEVEL_SECRET": "secret",
    "PROD_SECRET": None,
    "CONFIG_SECRET": "secret",
    }
    
    def _print_config(d, config):
        for key, value in config.items():
            if key in d:
                print(f"{key} = {value}")

    def _test(d, config):
        for key, value in d.items():
            if value is not None:
                assert config.get(key) == value, f"Expected {key} to be {value}, but got {config.get(key)}"
            else:
                assert config.get(key) is None, f"Expected {key} to be None, but got {config.get(key)}"


    config = get_config_tree(t2d/'app', deploy_name='devel')
    _test(d, config)
    
    d = {
        "DEVEL": "",
        "PRODUCTION": '1',
        "CONFIG": "CONFIG",
        "DEVEL_SECRET": None,
        "PROD_SECRET": "prod_secret",
        "CONFIG_SECRET": "secret",
    }

    config = get_config_tree(t2d/'app', deploy_name='prod')
    _test(d, config)


import jwt
import base64
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad
from Crypto.Random import get_random_bytes
from datetime import datetime, timedelta, timezone

def encrypt_token(token, encryption_key):
    cipher = AES.new(encryption_key, AES.MODE_CBC)
    iv = cipher.iv
    encrypted_token = cipher.encrypt(pad(token.encode(), AES.block_size))
    return base64.b64encode(iv + encrypted_token).decode()

def generate_encrypted_jwt(token, secret_key, encryption_key):
    # Encrypt the token
    encrypted_token = encrypt_token(token, encryption_key)

    # Create JWT payload with encrypted token and expiration
    payload = {
        'data': encrypted_token,
        'exp': datetime.now(timezone.utc) + timedelta(hours=1)  # Set expiration to 1 hour
    }

    # Sign the JWT using the secret key
    jwt_token = jwt.encode(payload, secret_key, algorithm='HS256')
    return jwt_token



def decrypt_token(encrypted_token_b64, encryption_key):
    encrypted_token_bytes = base64.b64decode(encrypted_token_b64)
    iv = encrypted_token_bytes[:16]  # Extract the IV (first 16 bytes)
    encrypted_token = encrypted_token_bytes[16:]  # Encrypted message after the IV
    cipher = AES.new(encryption_key, AES.MODE_CBC, iv)
    decrypted_token = unpad(cipher.decrypt(encrypted_token), AES.block_size)
    return decrypted_token.decode()

def verify_and_decrypt_jwt(jwt_token, secret_key, encryption_key):
    try:
        # Verify the JWT signature and get the payload
        payload = jwt.decode(jwt_token, secret_key, algorithms=['HS256'])

        # Extract the encrypted token from the payload
        encrypted_token = payload['data']

        # Decrypt the encrypted token
        decrypted_token = decrypt_token(encrypted_token, encryption_key)
        return decrypted_token
    except jwt.ExpiredSignatureError:
        return "Token has expired", 401
    except jwt.InvalidTokenError:
        return "Invalid token", 401


